package tester;

import java.lang.reflect.*;
import java.util.*;
import interfaces.*;

/**
 * This class illustrates the use of the Java reflection classes
 * and uses them to implement comparison for extenstional equality
 * between two objects of arbitrary type.
 * 
 * The comparison of circularly referential data will fail. 
 * To avoid this, the circularly referential class hierarchy 
 * must implement ISame interface that defines both the 
 * <code>same</code> method to compare two objects, and 
 * <code>makeString</code> method that constructs a <code>String</code>
 * representation of the object value.
 * 
 * @author Viera K. Proulx
 * @since 3 June 2007     
 *
 */
public class Inspector{

  /** the object to examine */
  Object sample;
  
  /** the tolerance for comparison of relative difference of inexact numbers */
  public static double TOLERANCE = 0.001;

  /** currnt indentation level for pretty-printing */
  public static String INDENT = "  ";

  /** 
   * Constructor: For the given instance get its <code>Class</code> 
   * and fields. 
   */
  Inspector(Object sample){
    this.sample = sample;
  }
  

  /**------------ THE METHODS USED TO COMPARE TWO OBJECTS -------------*/
  /**
   * <P>Compare the two given objects for extensional equality.<P>
   * <P>Consider inexact numbers (types <code>double</code>, 
   * <code>float</code>, <code>Double</code>, <code>Float</code>) 
   * to be the same if the relative difference
   * is below <code>TOLERANCE</code></P>
   * <P>Use <code>==</code> for all other primitive types 
   * and their wrapper classes.</P>
   * <P>Use the <code>String equals</code> method to compare two
   * objects of the type <code>String</code>.
   * <P>Traverse over <code>Arrays</code>, datasets that implement 
   * <CODE>{@link Traversal Traversal}</CODE> and datasets that implement
   * <code>Iterable</code>interface.</P>
   * <P>For the instances of a class that implements the 
   * <CODE>{@link ISame ISame}</CODE> interface, use its 
   * <CODE>same</CODE> method.</P>
   * <P>Special tests allow the user to provide the method to invoke
   * together with the argument list</P>
   * <P>Additionally, user may specify that a method invocation must 
   * throw a given <code>Exception</code> with the given message.</P>
   * 
   * @param obj1
   * @param obj2
   * @return true if the two given object are the same
   */
  public static boolean same(Object obj1, Object obj2){
    Reflector r1 = new Reflector(obj1);
    Reflector r2 = new Reflector(obj2);

    /** make sure both objects are not null */
    if (r1.sample == null)
      return r2.sample == null;
    if (r2.sample == null)
      return false;
    
    boolean sameClass = r1.sampleClass.equals(r2.sampleClass);
    
    if (sameClass){
 
      /** handle String objects separately */
      if (r1.sampleClass.getName().equals("java.lang.String")){
        return obj1.equals(obj2);
      }

      /** handle Color objects separately */
      if (r1.sampleClass.getName().equals("java.awt.Color") ||
          r1.sampleClass.getName().equals("colors.AColor") ||
          r1.sampleClass.getName().equals("colors.Red") ||
          r1.sampleClass.getName().equals("colors.Blue") ||
          r1.sampleClass.getName().equals("colors.Green") ||
          r1.sampleClass.getName().equals("colors.Yellow") ||
          r1.sampleClass.getName().equals("colors.White") ||
          r1.sampleClass.getName().equals("colors.Black") ){
        return ((java.awt.Color)obj1).getRGB() ==
               ((java.awt.Color)obj2).getRGB();
      }

      /** handle the primitive types separately */
      if (r1.sampleClass.isPrimitive()){
        if (isDouble(r1.sampleClass.getName()))
          return sameDouble((Double)obj1, (Double)obj2);
        else if (isFloat(r1.sampleClass.getName()))
          return sameFloat((Float)obj1, (Float)obj2);
        else
          return (obj1.equals(obj2));
      }

      /** handle the wrapper types separately */
      if (isWrapperClass(r1.sampleClass.getName())){
        if (isDouble(r1.sampleClass.getName()))
          return sameDouble((Double)obj1, (Double)obj2);
        else if (isFloat(r1.sampleClass.getName()))
          return sameFloat((Float)obj1, (Float)obj2);
        else
          return (obj1.equals(obj2));
      }

      /** handle Array objects */
      if ((obj1 instanceof Object[]) && (obj2 instanceof Object[])){
        int length = Array.getLength(obj1);
        if (Array.getLength(obj2) == length){
          for(int i = 0; i < length; i++){
            if (!same(((Object[])obj1)[i], ((Object[])obj2)[i]))
              return false;
          }
          return true;
        }
        else
          return false;
      }
            
      /** handle ISame objects by delegating to the user-defined method */
      if ((obj1 instanceof ISame) && (obj2 instanceof ISame))
        return ((ISame)obj1).same((ISame)obj2);
        
      
      /** handle Iterable objects */
      if ((obj1 instanceof Iterable) && (obj2 instanceof Iterable))
        return sameIterable((Iterable)obj1, (Iterable)obj2);

      /** now handle the general case */
      boolean sameValues = true;
      int i = 0;
      try{
        // this was used to verify the class names used in comparison
        // System.out.println("class 1: " + r1.sampleClass.getName());
        // System.out.println("class 2: " + r2.sampleClass.getName());
        for ( ; i < Array.getLength(r1.sampleDeclaredFields); i++){
          sameValues = sameValues && 
            same(r1.sampleDeclaredFields[i].get(obj1),
                r2.sampleDeclaredFields[i].get(obj2));
        }
      }

      catch(IllegalAccessException e){
        System.out.println("same comparing " + 
            r1.sampleDeclaredFields[i].getType().getName() + " and " +
            r2.sampleDeclaredFields[i].getType().getName() + 
            "cannot access the field " + i + 
            " message: " + e.getMessage());
        System.out.println("class 1: " + r1.sampleClass.getName());
        System.out.println("class 2: " + r2.sampleClass.getName());
      }

      return sameValues;
    }
    else
      return false;
  }

  /**
   * Determine whether the relative difference between two double numbers
   * is below the expected <code>TOLERANCE</code>
   * @param d1 the first inexact number
   * @param d2 the second inexact number
   * @return true is the two numbers are nearly the same 
   */
  public static boolean sameDouble(double d1, double d2){
    if (d1 - d2 == 0.0)
      return true;
    else{
     return (Math.abs(d1 - d2) / (Math.max (Math.abs(d1), Math.abs(d2))))
             < TOLERANCE;
    }
  }
 
  /**
   * Determine whether the relative difference between two float numbers
   * is below the expected <code>TOLERANCE</code>
   * @param d1 the first inexact number
   * @param d2 the second inexact number
   * @return true is the two numbers are nearly the same 
   */
  public static boolean sameFloat(float f1, float f2){
    if (f1 - f2 == 0.0)
      return true;
    else{
      Double d1 = ((Float)f1).doubleValue();
      Double d2 = ((Float)f2).doubleValue();
     return (Math.abs(d1 - d2) / (Math.max (Math.abs(d1), Math.abs(d2))))
             < TOLERANCE;
    }
  }

  /**
   * Determine whether two <code>Iterable</code> objects generate the same
   * data elements in the same order.
   * @param obj1 the first <code>Iterable</code> dataset
   * @param obj2 the second <code>Iterable</code> dataset
   * @return true is the two datasets are extensionally equal 
   */
  public static boolean sameIterable(Iterable obj1, Iterable obj2){
    Iterator it1 = obj1.iterator();
    Iterator it2 = obj2.iterator();
    
    return sameData(it1, it2);
  }
 
  /**
   * Determine whether two <code>Iterator</code>s generate the same data
   * in the same order.
   * @param it1 the first <code>Iterator</code>
   * @param it2 the second <code>Iterator</code>
   * @return true is both datasets containd the same data elements 
   * (in the same order)
   */
  public static boolean sameData(Iterator it1, Iterator it2){
    /** if the first dataset is empty, the second one has to be too */
    if (!it1.hasNext()){
      return !it2.hasNext();
    }
    /** the first dataset is nonempty - make sure the second one is too... */
    else if(!it2.hasNext())
      return false;
    /** now both have data - compare the next pair of data and recur */
    else{
      return same(it1.next(), it2.next()) &&
             sameData(it1, it2);
    }
  }

  
  /**------- THE METHODS USED TO DETERMINE THE TYPES OF OBJECTS ----------*/ 
  /**
   * Does the class with the given name represent inexact numbers?
   * @param name the name of the class in question
   * @return true if this is double, float, or Double or Float
   */
  public static boolean isDouble(String name){
    return 
    name.equals("double") ||
    name.equals("java.lang.Double");
  }
  
  /**
   * Does the class with the given name represent inexact numbers?
   * @param name the name of the class in question
   * @return true if this is double, float, or Double or Float
   */
  public static boolean isFloat(String name){
    return 
    name.equals("float") ||
    name.equals("java.lang.Float");
  }
  /**
   * Does the class with the given name represent a wrapper class for 
   * a primitive class?
   * @param name the name of the class in question
   * @return true if this is a class that represents a wrapper class
   */
  public static boolean isWrapperClass(String name){
    return 
    name.equals("java.lang.Integer") ||
    name.equals("java.lang.Long") ||
    name.equals("java.lang.Short") ||
    name.equals("java.math.BigInteger") ||
    name.equals("java.math.BigDecimal") ||
    name.equals("java.lang.Float") ||
    name.equals("java.lang.Double") ||
    name.equals("java.lang.Byte") ||
    name.equals("java.lang.Boolean") ||
    name.equals("java.lang.Character");           
  }
  
 
  /**
   * Produce a String representation of the given object.
   * <P>Show <code>String</code> 'as is'.</P>
   * <P>For primitive datatypes and thier wrapper classes show the 
   * primitive values.</P>
   * <P>For all <code>Array</code>s traverse over all elements.</P>
   * <P>For datatypes that implement <code>Iterable</code> interface
   *  traverse over all elements.</P>
   * <P>For datatypes that implement <CODE>{@link Traversal Traversal}</CODE>
   *  interface traverse over all elements.</P>
   *  <P>For an instance of a declared class show all fields</P>
   * 
   * @param obj the given object
   */
  public static String makeString(Object obj){

    // if the object is a String already - show it
    if (obj instanceof java.lang.String)
      return "\"" + obj.toString() +"\"";

    // if the object is a Color object already - show it
    if (obj instanceof java.awt.Color)
      return "\"" + obj.toString() +"\"";

    Class objClass = obj.getClass();
    
    // if the object is of primitive data type 
    // or an instance of a wrapper class - use default toString method 
    if (objClass.isPrimitive() ||
        isWrapperClass(objClass.getName())){
      return obj.toString(); 
    }
    
    // if the object is an Array - 
    // traverse over the data  
    if (obj instanceof Object[]){
      String result = "\n" + INDENT + " new Object[]()";
      INDENT = INDENT + "  ";
      for (int i = 0; i < Array.getLength(obj); i++){
        result = result + "){" + 
        makeString(((Object[])obj)[i]) + ",";
      }
      if (Array.getLength(obj) > 0){
        result = result.substring(0, result.length() - 1);
      }
      INDENT = INDENT.substring(0, INDENT.length()-2);
      return result = result + "}";
    }
    
    // for an instance of a declared class start with the class name 
    String result = "\n" + INDENT + " new " + objClass.getName() + "(";
    INDENT = INDENT + "  ";
    String field;
    
    // if the object is Iterable - 
    // traverse over the data generated by the iterator 
    if (obj instanceof Iterable){
      result = result + "){" + 
                 makeIterableStrings(((Iterable)obj).iterator()) + "}";
    }
    
    /** instance of  a class that may have several defined fields */
    else{
      Reflector r = new Reflector(obj);
      
      /** display all fields */
      for (Field f: r.sampleDeclaredFields){
        try{
          f.setAccessible(true);
          if ((f.get(obj)) == null)
            field = "this." + f.getName() + " = null";
          else
            field = "this." + f.getName() + " = " + makeString(f.get(obj)); 
          result = result + "\n" + INDENT + field;
          // to be tested - stops producing a string if it is too long
          if (result.length() > 50000)
            return result;
        }
        catch(IllegalAccessException e){
          System.out.println(
              "makeString cannot access the field " + f.getName() + 
              " of the class " + r.sampleClass.getName() +
              "\n   message: " + e.getMessage());
        }
      }
      /** close parentheses and finish up */
      result = result + ")";
    }
 
    INDENT = INDENT.substring(0, INDENT.length()-2);
    return result;
  }
  
  /**
   * Produce a <code>String</code> that represents the data generated
   * by the given iterator -- comma separated.
   * @param it the iterator for generating data
   * @return the <code>String</code> that represents all generated data
   */
  public static String makeIterableStrings(Iterator it){
    String result = "";
    while (it.hasNext()){
      result = result + makeString(it.next()) + ",";
    }
    /** remove the last comma - if any data is present */
    if (result.length() > 0)
      return result.substring(0, result.length()-1);
    else
      return result;
  }
}